(in-package :araneida)

;;; XXX Should abstract this somehow so that it can be done on things
;;; other than filename suffix

(defvar *content-types*
  '(("html" "text/html")
    ("gif" "image/gif")
    ("jpg" "image/jpeg")
    ("png" "image/png")
    ("class" "application/octet-stream")
    ("doc" "application/octet-stream")
    ("zip" "application/octet-stream")
    ("gz" "application/octet-stream")
    ("ASF" "video/x-ms-asf")
    ("tar" "application/octet-stream")
    ("avi" "video/x-msvideo")
    ("txt" "text/plain")))

(defun read-mime-types (filename)
  "Read a standard-format mime.types file and return an alist suitable for
assigning to *content-types*"
  (labels ((chop-comment (string)
                         (subseq string 0 (position #\# string)))
           (collect-extns (type extns)
                          (loop for e in (split extns)
				if (> (length e) 0)
                                collect (list e type))))
    (with-open-file (in filename :direction :input)
      (let ((mime-types (list))
            (eof (gensym)))
        (loop for line = (read-line in nil eof)
              until (eq line eof)
              for (type extns) =  (remove-if-empty
				   (split (chop-comment line) 2))
              append (collect-extns type extns))))))

(defun copy-stream (from to)
  "Copy into TO from FROM until end of the input file.  Do not
translate or otherwise maul anything. "
  (unless (eq (stream-element-type from) (stream-element-type to))
    (error "Incompatible streams ~A and ~A." from to))
  (let ((buf (make-array 4096 :element-type (stream-element-type from)))
        (since-last-pause 0))
    (do ((pos (read-sequence buf from) (read-sequence buf from)))
        ((= 0 pos) nil)      
      (write-sequence buf to :end pos))))

(defun send-file (r file-name &key content-type)
  (let ((stream (request-stream r))
        (if-mod-since (request-if-modified-since r 0)))
    (with-open-file (in file-name :direction :input)
      (peek-char nil in )               ;see if it really is readable
      (cond ((< if-mod-since (file-write-date in))
             (request-send-headers r :content-type content-type
                                   :content-length (file-length in)
                                   :last-modified (file-write-date in))
             (copy-stream in stream))
            (t
             (request-send-headers
              r :content-type content-type
	      :response-code 304 :response-text "Not modified"))))))


(defclass static-file-handler (handler)
  ((pathname :initarg :pathname :accessor static-file-pathname
	     :documentation "Root pathname for URI components to merge against")))

(defmethod handle-request-response
    ((handler static-file-handler) method request)
  ;; chop arg-string into /-delimited components.
  ;; remove .. components along with the component preceding them
  (let* ((path (cons :relative
		     (loop for p on
			   (nreverse (split
				      (request-unhandled-part request)
				      nil "/"))
			   if (string-equal (car p) "..")
			   do (setf p (cdr p))
			   else collect (car p) into v
			   finally (return (nreverse v)))))
	 (name (let ((n  (car (last path))))
		 (if (> (length n) 0) n nil)))
	 (path (butlast path))
	 (dot-pos (and name (position #\. name :from-end t)))
	 (extension (and dot-pos (subseq name (1+ dot-pos))))
	 (name (if dot-pos (subseq name 0  dot-pos) name))
	 (file (make-pathname :name name :directory path :type extension))
	 (fnam (merge-pathnames file (static-file-pathname handler)))
	 (content-type
	  (cadr (assoc (or (pathname-type fnam) "txt") *content-types*
		       :test #'string=))))
    (if (and (not (wild-pathname-p fnam)) (probe-file fnam))
        (with-open-file (in fnam :direction :input)
          (when (ignore-errors (peek-char nil in nil))
	    (send-file request fnam :content-type content-type)
	    t))
	(values nil (format nil "Can't find file ~S~%" fnam)))))

#|
(install-handler *root-handler* (make-instance 'static-file-handler :pathname #p"/tmp/fooo")  "http://ww.noetbook.telent.net/tmp/logo" nil)
|#
